/*----------------------------------------------------------------------
	File Name	: dcache.c
	compiler 	: gcc
	Edited by	: Lee, WoonJae
	History 	: 1995.8.24 -- ο TRIE  
	Last Change	: 1996.7.24 --  ħ
			: 1997.5.20 -- csfive(linux)
				segmentation fault߻
			: 1999.5.26 -- TDBM data size ũ
			: 1999.7.27 -- trie.c dcache.c и

	cache page Ϸ  Ͽ DC_MAX_PAGE ã
	ġ ȴ.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include <sys/types.h>  /* for low level file operation */
#include <sys/stat.h>


#ifdef _MSC_VER		/* SYSTEM dependent define */
#include <io.h>
#else
#include <sys/uio.h>
#include <unistd.h>
#endif

#include <fcntl.h>
#include <errno.h>

/* Debugging options */
/*
#define HITINFO 1
*/

#include "dcache.h"
#include "global.h"

#define DEFAULT_FMODE	0644

/* ּ    
 ּ Ÿ(idx) page offset ȴ.
    MKIDX() : page offset Ͽ index(Ǵ address) .
    PMASK() : index page Ѵ.
    OMASK() : index offset Ѵ.
*/

#define MKIDX(p,o) (((p) << 8) | (o))
#define PMASK(i) ((i) >> 8)
#define OMASK(i) ((i) & 0x0ff)

#define D_POS(dc,dpage) ((dpage)*dc_pagesize(dc))
#define C_POS(dc,cpage) ((dc)->cache_buffer+(cpage)*dc_pagesize(dc))
        /*  : *cache_buffer size  ٸ  ִ.*/

/* 
	page_cmp() bsearch() ȴ.
*/
int page_cmp(DC_PAGE *a, DC_PAGE *b)
{
	return (int)(*a - *b);
}

/*
	bsearch ϱ Ͽ dc_table sort · Ѵ.
	dct[idx] شϴ ׸  ġ ű ϸ Ѵ.
*/
void dct_sort(DCTYPE dct[], DC_IDX idx, DC_PAGE cpage_end)
{
	DCTYPE tmpdc;
	
	tmpdc = dct[idx];
	for(;idx < cpage_end-1 && tmpdc.dpage > dct[idx+1].dpage;idx++)
		dct[idx] = dct[idx+1];
	for(;idx > 0 && tmpdc.dpage < dct[idx-1].dpage;idx--)
		dct[idx] = dct[idx-1];
	dct[idx] = tmpdc;
}

/* lru_update() LRU table ϱ Ͽ Ǹ
	 ֱٿ   head ű
	     tail ǥѴ.
 :
	prev next NULL  ˻ ʴ´.
	lru_head lru_tail  check ϴٰ .
*/
void lru_update(DCACHE *dc, DC_PAGE cpage)
{
    	CDTYPE *cdt;
    	DC_PAGE p,n;
	
	if(cpage == dc->lru_head)
		return;

	cdt=dc->cd_table;

	/* prev next Ѵ */
	p = cdt[cpage].prev_cpage;
	n = cdt[cpage].next_cpage;

	/* t link  prev next  */
	cdt[p].next_cpage = n;		/* next prev Ѵ. */
	if(cpage == dc->lru_tail) 		/* t ̾ٸ */
		dc->lru_tail = p;	/* prev  ȴ. */
	else			
		cdt[n].prev_cpage = p;	/* next  
					   next prev link */

	/* t lru head ġ ´. */
	cdt[dc->lru_head].prev_cpage = cpage;
	cdt[cpage].next_cpage = dc->lru_head;
	dc->lru_head = cpage;
}


/*
	Ͽ  page а  Լ
*/
void dc_write(DCACHE *dc, DC_PAGE dpage, DC_PAGE cpage) 
{
    	int fpos;
    	if(dc->cd_table[cpage].dirty)
	{
	    fpos=lseek(dc->fid,D_POS(dc,dpage),SEEK_SET);
	    write(dc->fid,C_POS(dc,cpage),dc_pagesize(dc));
	    if(fpos > dc->fend)
		dc->fend=fpos;
	}
    	dc->cd_table[cpage].dirty=0;
}
void dc_read(DCACHE *dc, DC_PAGE dpage, DC_PAGE cpage) 
{
    	if((int)D_POS(dc,dpage) <= dc->fend)
	{
	    lseek(dc->fid,D_POS(dc,dpage),SEEK_SET);
	    read(dc->fid,C_POS(dc,cpage),dc_pagesize(dc));
	}
    	dc->cd_table[cpage].dirty=0;
}

/*
	getpage()
	ϴ  cache ã ش.
	cache  Ͽ о cache ø.
*/

DC_PAGE getpage(DCACHE *dc, DC_PAGE dpage)
{
	DCTYPE	*dct;
	CDTYPE	*cdt;
	DC_PAGE lru_tail;
	DC_PAGE	c_end;
	DC_PAGE	cpage;

	DCTYPE	*dc_pages;
	
	/* ʱⰪ  */
	dct = dc->dc_table;
	cdt = dc->cd_table;
	c_end = dc->cache_end;
	lru_tail = dc->lru_tail;

	if(dc->lru_head < DC_MAX_PAGE && cdt[dc->lru_head].dpage == dpage)
	{
		return dc->lru_head;
		/* head  ׳ Ѱش. */
	}

	/* cache table ã´ */
	dc_pages =  (DCTYPE *)bsearch(&dpage, dct, c_end, sizeof(DCTYPE),
			(int (*)(const void *,const void *))page_cmp);

	if(dc_pages != NULL) /*  */
	{
#ifdef HITINFO
dc->chit++;
#endif
		cpage = dc_pages->cpage;
		lru_update(dc, cpage);
		return cpage;
	}
	else /*  */
	{
#ifdef HITINFO
dc->cfail++;
#endif
		if(dc->cache_end < DC_MAX_PAGE)
		{
			/* cache       */

			cpage = dc->cache_end;
			dc_read(dc,dpage,cpage);

			/* new lru && new cdt */
			cdt[cpage].next_cpage = dc->lru_head;
			cdt[cpage].prev_cpage = DC_MAX_PAGE;
			if(dc->lru_head < DC_MAX_PAGE)
				cdt[dc->lru_head].prev_cpage = cpage;
			dc->lru_head = cpage;
			if(dc->lru_tail >= DC_MAX_PAGE)
				dc->lru_tail = cpage;
			cdt[cpage].dpage = dpage;

			/* new dct .. */
			dct[cpage].dpage = dpage;
			dct[cpage].cpage = cpage;
			dc->cache_end++;
			dct_sort(dct,cpage,dc->cache_end);
		}
		else
		{
			/* cache update */
		    	DC_PAGE tpage;

			cpage = dc->lru_tail;
			lru_update(dc, cpage);

			tpage = cdt[cpage].dpage;
			dc_write(dc,tpage,cpage);
			dc_read(dc,dpage,cpage);

			dc_pages = 
			(DCTYPE *)bsearch(&tpage, dct, c_end, sizeof(DCTYPE),
				(int (*)(const void *,const void *))page_cmp);

			dc_pages->dpage = dpage;
			cdt[cpage].dpage = dpage;

			dct_sort(dct,(DC_IDX)(dc_pages - dct),dc->cache_end);
		}
		return cpage;			
	}
}

char *dc_add_r(DCACHE *dc, DC_IDX idx)
{
	return C_POS(dc,getpage(dc,PMASK(idx)))+OMASK(idx)*(dc->unitsize);
}
char *dc_add_w(DCACHE *dc, DC_IDX idx)
{
    	DC_PAGE cpage;
	cpage=getpage(dc,PMASK(idx));
	if((dc->flag & O_RDWR) == 0)
	{
		fprintf(stderr,"%s:%s\n",__FILE__,"check open flags error");
		exit(0);
	}
	dc->cd_table[cpage].dirty=1;
	return C_POS(dc,cpage)+OMASK(idx)*(dc->unitsize);
}

void dc_dirty_set(DCACHE *dc, char *x)
{
    	DC_PAGE cpage;
	if((dc->flag & O_RDWR) == 0)
	{
		fprintf(stderr,"%s:%s\n",__FILE__,"check open flags error");
		exit(0);
	}
	cpage=(x-dc->cache_buffer)/dc_pagesize(dc);
	dc->cd_table[cpage].dirty=1;
}

DCACHE * dcache_open(char *fname, int flag, DC_IDX unitsize)
{
	DCACHE *dc;

	dc = (DCACHE *)malloc(sizeof(DCACHE));
	dc->unitsize=unitsize;
	dc->fname = (char *)malloc(strlen(fname)+1);
	strcpy(dc->fname,fname);
	dc->flag = flag;

	if((dc->fid = open(fname,flag,DEFAULT_FMODE)) == -1)
	{
		fprintf(stderr,"dcache_open:ERROR: can not open file '%s'\n",fname);
		exit(0);
	}

	dc->fend=lseek(dc->fid,0,SEEK_END);
	dc->cache_end = 0;
	dc->cache_buffer = (char *)malloc(dc_pagesize(dc)*DC_MAX_PAGE);
	dc->dc_table = (DCTYPE *)malloc(sizeof(DCTYPE)*DC_MAX_PAGE);
	dc->cd_table = (CDTYPE *)malloc(sizeof(CDTYPE)*DC_MAX_PAGE);
	dc->lru_head = DC_MAX_PAGE;
	dc->lru_tail = DC_MAX_PAGE;
	dc->chit=0; dc->cfail=0;

	return dc;
}


/*
   	dcache_close
	PUBLIC
*/

void dcache_close(DCACHE *dc)
{
#ifdef HITINFO
	/* stat */
	{
		float f;
		fprintf(stderr,"dcache_close:statistics:'%s' \n",dc->fname);
		f = (float)dc->chit/((float)dc->chit+(float)dc->cfail);
		fprintf(stderr,"cache hit ratio = ");
		fprintf(stderr,"%f%% total %d access\n",f*100,dc->chit+dc->cfail);
		fprintf(stderr,"size of unit = %d, ",dc->unitsize);
		fprintf(stderr,"pagesize = %d\n",DC_PAGESIZE);
		fprintf(stderr,"size of cache buffer = %dbyte\n",
			dc->unitsize*DC_PAGESIZE*DC_MAX_PAGE);
	}
#endif

	if(dc->flag & O_RDWR)
	{
	    	DC_PAGE cpage;
		CDTYPE *cdt;

		cdt = dc->cd_table;
		for(cpage=0;cpage<dc->cache_end;cpage++)
			dc_write(dc, cdt[cpage].dpage, cpage);
	}

	free(dc->cache_buffer);
	free(dc->cd_table);
	free(dc->dc_table);
	free(dc->fname);
	close(dc->fid);
	free(dc);
}
void dcache_truncate(DCACHE *dc, DC_IDX size)
{
#ifndef _MSC_VER
	if(dc->flag & O_RDWR)
	    ftruncate(dc->fid,size);
#endif
}

