#include <stdio.h>
#include <fcntl.h>

#include "simti.h"

/*----------------------------------------------------------------------
	_init(), _read(), _write(), _close()
----------------------------------------------------------------------*/
SIMTI * simti_init()
{
	SIMTI *stp;

	stp = (SIMTI *)malloc(sizeof(SIMTI));
	stp->filename = (char *)malloc(10);
	strcpy(stp->filename , "noname");
	stp->flag = 0;
	stp->change = 0; 
	stp->search_end=0;
	stp->head = (HEADI *)malloc(sizeof(HEADI));
	stp->nf = (ST_NF *)malloc(sizeof(ST_NF) * ST_NF_DEFAULT);

	strcpy(stp->head->version,ST_VER);
	stp->head->n_size = ST_NF_DEFAULT;
	stp->head->f_size = ST_NF_DEFAULT-1;

	stp->head->s_node.K = 0;
	stp->head->s_node.CS = 0;
	stp->head->s_node.I = 0;
	stp->head->s_node.child = 0;

	stp->head->s_free.size = 0;
	stp->head->s_free.next = 1; 

	/* 0    */
	stp->nf[1].free.size = ST_NF_DEFAULT-1; 
	stp->nf[1].free.next = 0; 
	return stp;
}

SIMTI *simti_clean(SIMTI *stp)
{
	stp->flag = 0;
	stp->change = 0; 
	stp->search_end=0;

	stp->head->n_size = ST_NF_DEFAULT;
	stp->head->f_size = ST_NF_DEFAULT-1;

	stp->head->s_node.K = 0;
	stp->head->s_node.CS = 0;
	stp->head->s_node.I = 0;
	stp->head->s_node.child = 0;

	stp->head->s_free.size = 0;
	stp->head->s_free.next = 1; 

	/* 0    */
	stp->nf[1].free.size = ST_NF_DEFAULT-1; 
	stp->nf[1].free.next = 0; 
	return stp;
}

SIMTI * simti_read(char *fname, char flag)
{
	int fid;
	SIMTI *stp;
	unsigned long n_size, f_size, next;

	if((fid = open(fname,O_RDONLY)) == -1)
		return NULL;

	stp = (SIMTI *)malloc(sizeof(SIMTI));
	stp->filename = (char *)malloc(strlen(fname)+1);
	strcpy(stp->filename,fname);
	stp->flag = flag;
	stp->change = 0;
	stp->search_end=0;

	stp->head = (HEADI *)malloc(sizeof(HEADI));
	read(fid,stp->head,sizeof(HEADI));

	if(stp->flag=='w')
	{
		n_size=stp->head->n_size+ST_NF_DEFAULT;
		f_size=stp->head->f_size+ST_NF_DEFAULT;
	}
	else /* r */
	{
		n_size=stp->head->n_size;
		f_size=stp->head->f_size;
	}
	stp->nf = (ST_NF *)malloc(sizeof(ST_NF) * n_size);
	read(fid,stp->nf,sizeof(ST_NF) * stp->head->n_size);
	if(stp->flag=='w')
	{
		next=stp->head->s_free.next;
		if(next)
		{
			while(stp->nf[next].free.next)
				next=stp->nf[next].free.next;
			if(next + stp->nf[next].free.size == stp->head->n_size)
				stp->nf[next].free.size+=ST_NF_DEFAULT;
			else
			{
				stp->nf[next].free.next=stp->head->n_size;
				next= stp->nf[next].free.next;
				stp->nf[next].free.size=ST_NF_DEFAULT;
				stp->nf[next].free.next=0;
			}
		}
		else
		{
			stp->nf[stp->head->n_size].free.size=ST_NF_DEFAULT;
			stp->nf[stp->head->n_size].free.next=0;
			stp->head->s_free.next=stp->head->n_size;
		}
		stp->head->n_size=n_size;
		stp->head->f_size=f_size;
	}

	close(fid);
	return stp;
}

int 	simti_write(SIMTI *stp, char *fname)
{
	int fid;
	ST_FREE *freep;

/*
	if(stp->change==0)
	{
		/*    * /
		return ;
	}
*/
	if(stp!=NULL && strcmp(stp->filename,fname) ==0)
	{
		if(stp->flag == 'r')
			return -1;
		fprintf(stderr," ̸  '%s'\n",fname);
	}
	else
		fprintf(stderr," ̸  '%s'\n",fname);
		
	if((fid = open(fname,O_RDWR|O_CREAT,DEFAULT_FMODE)) == -1)
	{
		fprintf(stderr,"simti_write ERROR: can't open file '%s'\n",fname);
		return -1;
	}

	freep=&stp->head->s_free;
	if(freep->next)
	{
		while(stp->nf[freep->next].free.next)
			freep=&stp->nf[freep->next].free;
		if(freep->next + stp->nf[freep->next].free.size == stp->head->n_size)
		{
			stp->head->n_size -= stp->nf[freep->next].free.size;
			stp->head->f_size -= stp->nf[freep->next].free.size;
			freep->next=0;
		}
	}
	write(fid,stp->head,sizeof(HEADI));
	write(fid,stp->nf,sizeof(ST_NF) * stp->head->n_size);
	close(fid);
}

int 	simti_close(SIMTI *stp)
{
	if(stp != NULL)
	{
		if(stp->filename != NULL)
			free(stp->filename);
		if(stp->head != NULL)
			free(stp->head);
		if(stp->nf != NULL)
			free(stp->nf);
	}
}

void simti_stat(SIMTI * stp)
{

	fprintf(stderr,"SIMTI state\n");
	fprintf(stderr,"\tname:%s\n",stp->filename);
	fprintf(stderr,"\tversion:%s\n",stp->head->version);
	fprintf(stderr,"\tflag:%c\n",stp->flag);
	fprintf(stderr,"\tchange:%c\n",stp->change);
	fprintf(stderr,"\tnsize:%d\n",stp->head->n_size);
	fprintf(stderr,"\tfsize:%d\n",stp->head->f_size);
}
/*----------------------------------------------------------------------
	_alloc(), _free()
----------------------------------------------------------------------*/

int	simti_free(SIMTI *stp, INDEX idx, unsigned long size)
{	/* idx  free  index */
	/* size is free   */

	INDEX	i,prev_i=0;
	ST_FREE	*start;

	if(size <=0)
		return -1;
	if(idx <=0 || idx+size>=stp->head->n_size)
		return -1;
	
	i=stp->head->s_free.next;

	if (i==0)		/* no free node */
		{
		stp->head->s_free.next=idx;
		stp->nf[idx].free.size=size;
		stp->nf[idx].free.next=0;
		return 0;
		}

	if ( idx < i )		/* idx is the smallest in free */
		{
		stp->head->s_free.next=idx;
		if (i == idx + size ) 
			{
			stp->nf[idx].free.size=size + stp->nf[i].free.size;
			stp->nf[idx].free.next=stp->nf[i].free.next;
			}
		else 	{
			stp->nf[idx].free.size=size;
			stp->nf[idx].free.next=i;
			}
		stp->head->f_size += size;
		return 0;
		}

	while (i != 0 && i<idx)	/* other wise */
		{
		prev_i=i;
		i=stp->nf[i].free.next;
		}
	start=&(stp->nf[prev_i].free);	/* prev_i != 0 */
	
	if (idx+size == i)	/* ٷ  尡 free */
		{		
		size += stp->nf[i].free.size;
		start->next = stp->nf[i].free.next;
		stp->head->f_size -= stp->nf[i].free.size;
		}

	if (prev_i + start->size == idx )
		start->size += size;
	else 	{
		stp->nf[idx].free.size = size;
		stp->nf[idx].free.next = start->next;
		start->next = idx;
		}
	stp->head->f_size += size;
	return	0;
}

INDEX 	simti_alloc(SIMTI *stp, unsigned long size)
{
	INDEX 	i,prev_i=0;

	for (i=stp->head->s_free.next ; i!=0 ; i=stp->nf[i].free.next )
		{
		if ( stp->nf[i].free.size >= size ) break;
		prev_i=i;
		}

	if (i==0) 
	{
		fprintf(stderr,"simti_alloc:NO FREE NODE\n");
		return 0;
//		exit();
		/*return 0;	/* no free node */
	}
	if (prev_i==0)		/* head node */
		if (size == stp->nf[i].free.size) 
			stp->head->s_free.next=stp->nf[i].free.next;
		else 	{	/* size < free node size */
		 	stp->nf[i+size].free.size = stp->nf[i].free.size - size;
			stp->head->s_free.next = i+size;
			stp->nf[i+size].free.next = stp->nf[i].free.next;
			}
	else

		if (size == stp->nf[i].free.size) 
			stp->nf[prev_i].free.next=stp->nf[i].free.next;
		else	{
	 		stp->nf[i+size].free.size = stp->nf[i].free.size - size;
			stp->nf[prev_i].free.next = i+size;
			stp->nf[i+size].free.next = stp->nf[i].free.next;
			}
	stp->head->f_size -= size;
	return i;
}

/*----------------------------------------------------------------------
	_search(), _firstkey(), _nextkey()
----------------------------------------------------------------------*/

int kcomp(ST_NODE *a, ST_NODE *b)
{
	return (int)a->K - (int)b->K;
}

int simti_search(SIMTI *stp, unsigned char *word)
{
	int i;
	ST_NODE tmpnode,*rnode;
	unsigned long child,newidx;
	unsigned char cs;

	for(i = 0; *word && i < stp->search_end; i++) {
		if(*word == stp->search_word[i]) word++;
		else break;
	}
	stp->search_end = i;
	if(stp->search_end == 0 ) {
		cs = stp->head->s_node.CS;
		child = stp->head->s_node.child;
	}
	else {
		child = stp->search_idx[stp->search_end-1];
		cs = stp->nf[child].node.CS;
		child = stp->nf[child].node.child;
	}
	while(*word && cs != 0) {
		tmpnode.K=*word;
		rnode = (ST_NODE *)bsearch(&tmpnode, &stp->nf[child], cs, sizeof(ST_NODE), 
			(int (*)(const void *, const void *))kcomp) ;
		if(rnode == NULL) break;
		else {
			newidx= rnode - (ST_NODE *)stp->nf;
			stp->search_word[stp->search_end] = *word;
			stp->search_idx[stp->search_end] = newidx;
			stp->search_end++;
			word++;
			child = stp->nf[newidx].node.child;
			cs = stp->nf[newidx].node.CS;
		}
	}
	return stp->search_end;
}


int 	simti_firstkey(SIMTI *stp, unsigned char *word)
{
	int i;
	unsigned long index;
	unsigned char cs;

	index = stp->head->s_node.child;
	cs = stp->head->s_node.CS;

	i=0;
	while(cs) {
		word[i] = stp->search_word[i] = stp->nf[index].node.K;
		stp->search_idx[i]=index;
		i++;
		if(stp->nf[index].node.I) 
			break;
		cs = stp->nf[index].node.CS;	/* Ʒ    */
		index = stp->nf[index].node.child;
	}
	word[i]=0;
	return stp->search_end=i;
}

int 	simti_nextkey(SIMTI *stp, unsigned char *word)
{
	int i;
	unsigned long index;
	unsigned char cs;
	ST_NODE *parent;

	if(stp->search_end <= 0) 
		return 0;
	for(i = 0; i < stp->search_end; i++) 
		word[i] = stp->search_word[i];

	index = stp->search_idx[i - 1]; /* i search_end  . */
	if(i == 1)
		parent = &stp->head->s_node;
	else
		parent = &stp->nf[stp->search_idx[i - 2]].node;

	cs = stp->nf[index].node.CS;

	/* parent --> index --> child */
	/*            sibling         */
	while(i > 0) {
		if(cs) { /* child ִ  */
			parent = &stp->nf[index].node;
			index=stp->nf[index].node.child;
			cs = stp->nf[index].node.CS;

			word[i] = stp->search_word[i] = stp->nf[index].node.K;
			stp->search_idx[i] = index;
			i++;
			if(stp->nf[index].node.I) 
				break;
		}
		else if( index < parent->child + parent->CS -1) { /* sibling ִ  */
							/*   -1   */
			index++;
			cs = stp->nf[index].node.CS;

			word[i-1] = stp->search_word[i-1] = stp->nf[index].node.K;
			stp->search_idx[i-1] = index;
			if(stp->nf[index].node.I) 
				break;
		}
		else { /* child sibling   */
			i--;
			if(i<=0)
			{
				i=0;
				break;
			}
			index = stp->search_idx[i-1];
			if(i == 1)
				parent = &stp->head->s_node;
			else
				parent = &stp->nf[stp->search_idx[i-2]].node;
			cs = 0;
		}
	}
	word[i] = 0;
	return stp->search_end=i;
}

/*----------------------------------------------------------------------
	_insert(), _delete() : yuntan
----------------------------------------------------------------------*/


INDEX binary_search(SIMTI *stp, INDEX idx, unsigned char size,unsigned char key)
{
	unsigned long left=0,right,middle;
	unsigned char K;
	ST_NODE	*node;

	right=(unsigned long) size -1;
	while(left<=right)
		{
		middle=(left+right)/2;
		node=&(stp->nf[middle+idx].node);
		if (K > node->K)
			left=middle+1;
		else if (K < node->K)
			right=middle-1;
		else	return(idx+middle);
		}
	return 0;
}
			
int simti_search_2(SIMTI *stp, unsigned char *word)
{
	INDEX		current,target;
	unsigned char	size;
	int		len,i=0;
	
	if (stp->head->s_node.CS == 0)	return 0;

	i=0;
	while((word[i] != 0) && i<stp->search_end)
		{
		if (word[i]!=stp->search_word[i]) break;
		i++;
		}

	stp->search_end=i;

	if (stp->search_end==0)	
		{
		current=stp->head->s_node.child;
		size=stp->head->s_node.CS;
		}
	else	{
		current=stp->search_idx[stp->search_end];
		size=stp->nf[current].node.CS;
		}
		
	len=strlen(word);

	while(i<len)
		{				
		if (stp->nf[current].node.CS == 0 )	break;

		current=stp->nf[current].node.child;
		target=binary_search(stp,current,size,word[i]);
		if (target != 0)
			{
			stp->search_word[i]=word[i];
			stp->search_idx[i]=target;
			i++;
			}
		else break;
		}
		stp->search_end=i;
		return i;
}

int simti_insert2(SIMTI *stp, unsigned char *word, unsigned short I)
{
	int len,i,j,size,current;
	INDEX 	idx, new_child, old_child;
	INDEX	new[ST_MAX_WORD];
	ST_NODE	*parent;

	len=strlen(word);
		/* current should equal search_end */
	if (stp->head->s_node.CS == 0)	/* the first insertion */
		{
		for(i=0;i<len;i++)
			{
			new[i]=simti_alloc(stp,1);
			if (new[i]==0)	
				{	/* if failed to get new nodes, as possible as, free all */
				for(j=0;j<i;j++)
					simti_free(stp,j,1);
				return -1;
				}
			}
		for(i=0;i<len-1;i++)
			{
			stp->nf[new[i]].node.child=new[i+1];
			stp->nf[new[i]].node.I=0;
			stp->nf[new[i]].node.CS=1;
			stp->nf[new[i]].node.K=word[i];
			}
		stp->nf[new[i]].node.K=word[i];
		stp->nf[new[i]].node.child=0;
		stp->nf[new[i]].node.CS=0;
		stp->nf[new[i]].node.I=I;

		return 1;
		}
			

	if (stp->search_end==len) /* return 0;	/* duplicated */
		{
		if (stp->nf[len].node.I == I)	return 0;
		else	{
		/*	simti_newinfo(I,stp->nf[len].node.I);	/* new information needs 
			return 1;					*/
			return -1;
			}
		}
	
	idx=stp->search_idx[stp->search_end];		/* parent of target word */
	if (stp->search_end==0)
		parent=&(stp->head->s_node);
	else parent=&(stp->nf[idx].node);
	size = parent->CS;
	old_child= parent->child;	/* old child */

	new_child= simti_alloc(stp,size + 1);	/* make a siblings of new node */

	if (new_child==0)	return -1;	/* alloc of  new child is failed */

	for(i=0; i<size; i++)		/* insert a new node into new siblings */
		if (word[stp->search_end] > stp->nf[old_child+i].node.K) break;
		else 
			stp->nf[new_child+1].node=stp->nf[old_child+i].node;

	current=new_child+i;
	stp->nf[current].node.K=word[stp->search_end];

	while(i<size)
		{
		stp->nf[new_child+i+1].node= stp->nf[old_child+i].node;
		i++;
		}

	if (stp->search_end == len-1)		/* the last node is already inserted */
		{
		stp->nf[current].node.I=I;
		stp->nf[current].node.CS=0;
		stp->nf[current].node.child=0;
		}
	else
		{
		stp->nf[current].node.I=0;
		stp->nf[current].node.CS=1;

		for(i=0;i<len-stp->search_end;i++)
			{
			new[i]=simti_alloc(stp,1);
			if (new[i]==0)	
				{	/* if failed to get new nodes, as possible as, free all */
				for(j=0;j<i;j++)
					simti_free(stp,j,1);
				simti_free(stp,new_child,size+1);
				return -1;
				}
			}
		stp->nf[current].node.child=new[0];
		for(i=0;i<len-stp->search_end-1;i++)
			{
			current=new[i];
			stp->nf[current].node.child=new[i+1];
			stp->nf[current].node.I=0;
			stp->nf[current].node.CS=1;
			stp->nf[current].node.K=word[stp->search_end+i];
			}
		stp->nf[new[i]].node.K=word[len-1];
		stp->nf[new[i]].node.child=0;
		stp->nf[new[i]].node.CS=0;
		stp->nf[new[i]].node.I=I;
		}
	if  (simti_free(stp,old_child,size) == -1) return -1;

	parent->child=new_child;
	parent->CS=size+1;
	return 1;
}

int 	simti_insert(SIMTI *stp, unsigned char *word, unsigned short I)
{
	long child_index, new_index;
	int i,j;
	unsigned char cs;
	ST_NODE *parent, tmp_node = {0, 0, 0, 0};

	if(*word == 0) 
		return -1;
	
	simti_search(stp, word);
	word += stp->search_end;

	if(stp->search_end == 0) 
		parent = &stp->head->s_node;
	else 
		parent = &stp->nf[stp->search_idx[stp->search_end - 1]].node;

	while(*word) {
		cs = parent->CS;
		if(cs == 0) { /* child   */
			new_index = simti_alloc(stp, 1);
			stp->nf[new_index].node = tmp_node;
			stp->nf[new_index].node.K = *word;
			parent->CS = 1;
			parent->child = new_index;
			stp->search_idx[stp->search_end] = new_index;
			stp->search_word[stp->search_end] = *word;
			stp->search_end++;
			word++;
			parent = &stp->nf[new_index].node;
		}
		else {
			new_index = simti_alloc(stp, cs + 1);
			child_index = parent->child;
			for(i = 0; i < cs; i++) {
				if(stp->nf[child_index + i].node.K < *word) 
					stp->nf[new_index + i].node = stp->nf[child_index + i].node;
				else 
					break;
			}

			stp->nf[new_index + i].node = tmp_node;
			stp->nf[new_index + i].node.K = *word;

			stp->search_idx[stp->search_end] = new_index + i;
			stp->search_word[stp->search_end] = *word;
			stp->search_end++;
			word++;

			for(j = i; j < cs; j++) 
				stp->nf[new_index + j + 1].node = stp->nf[child_index + j].node;
			parent->child = new_index;
			parent->CS = cs + 1;
			simti_free(stp, child_index, cs);

			parent = &stp->nf[new_index + i].node;
		}
	}
	if(parent->I==0) {
		parent->I = I;
		return 1;
	}
	else
		return 0;
}

int 	simti_replace(SIMTI *stp, unsigned char *word, unsigned short I)
{
	if(*word == 0) 
		return -1;
	
	simti_search(stp, word);
	word += stp->search_end;

	if(stp->search_end == 0 || *word) 
		return -1;
	else 
	{
		stp->nf[stp->search_idx[stp->search_end - 1]].node.I = I;
		return 1;
	}
}

int	simti_lookup(SIMTI *stp, unsigned char *word, unsigned short *I_buffer)
{
	int	i;
	if (simti_search(stp,word)==0) 
		return 0;
	else
		for(i=0 ; i<stp->search_end ; i++)
			I_buffer[i]=stp->nf[stp->search_idx[i]].node.I;
	return stp->search_end;
}


unsigned short  simti_fetch(SIMTI *stp, unsigned char *word)
{
	simti_search(stp,word);
	if (*(word + stp->search_end) !=0)
		return 0;
	else return stp->nf[stp->search_idx[stp->search_end-1]].node.I;
}


int 	simti_delete(SIMTI *stp, unsigned char *word)
{
	int	i,d,j;
	INDEX	idx,new;
	unsigned char	size;
	ST_NODE	node,*temp;

	simti_search(stp,word);
	if (*(word + stp->search_end ) != 0 || *word == 0 )	
		return 0;	

	temp= &(stp->nf[stp->search_idx[stp->search_end-1]].node);

	if (temp->I == 0 ) return 0;

	node=*temp;

	for(i=stp->search_end-1;i>0 && node.CS==0 && node.I == 0;i--)
		{
		stp->search_end--;
		if (i==1) 
			node=stp->head->s_node;
		else node=stp->nf[stp->search_idx[i-1]].node;

		if (node.CS == 1)
			{
			simti_free(stp,node.child,1);
			node.CS= 0;
			node.child=0;
			}
		else{
			idx= node.child;
			d = stp->search_idx[i] - idx;	/* distance */
			size = node.CS;
		
			new=simti_alloc(stp,size-1);
			for(j=0;j<d;j++)
				stp->nf[new+j].node=stp->nf[idx+j].node;
			for(j=0;j<size-d-1;j++)
				stp->nf[new+j].node=stp->nf[idx+j].node;
			simti_free(stp,idx,size);
			node.CS--;
			node.child=new;
			}
		if (i==1) 
			stp->head->s_node=node;
		else	stp->nf[stp->search_idx[i-1]].node=node;
		}
	return 1;
}

