/*----------------------------------------------------------------------
	File Name	: trie.c
	compiler 	: gcc
	Edited by	: Lee, WoonJae
	History 	: 1995.8.24 -- ο TRIE  
	Last Change	: 1996.7.24 --  ħ
			: 1997.5.20 -- csfive(linux)
				segmentation fault߻
			: 1999.5.26 -- TDBM data size ũ
			  1999.7.29
	
	trie.h  INFO node  ޶ 
	 α׷ ĥ ʿ䰡 .
	DOS version	:     ĥ  .
	BUG		: INFO     ϴµ
				Word_ID . --ħ
----------------------------------------------------------------------*/
#include <stdio.h>		/* for stderr */
#include <string.h>		/* for strcpy */
#include <malloc.h>		/* for malloc */
#include <fcntl.h>		/* for O_RDWR|O_CREAT */
#include <stdlib.h>

#include "trie.h"
#include "global.h"

#ifdef _MSC_VER
#define FILE_FLAG_R (O_RDONLY|O_BINARY)
#define FILE_FLAG_W (O_RDWR|O_CREAT|O_BINARY)
#else
#define FILE_FLAG_R (O_RDONLY)
#define FILE_FLAG_W (O_RDWR|O_CREAT)
#endif


/*----------------------------------------
  <section> low level define	
  :hidden
  
  trie file head+node array ȴ.
  head ũ  pagesize*unitsize
  head = system head + user head
  node array 0 FREE_NODE pointer ְ
  		1 START_NODE(TNODE) pointer ִ.
----------------------------------------------------------------------*/
#define	FREE_NODE 0
#define START_NODE 1

/*----------------------------------------
  <section> dcache interface
  :hidden, user ϸ  ʴ  .
  
  TNODE,      FREE        INFO
  getnode(),  getfnode(), getinode()	: idx  memory pointer
  getnaddr(), getfaddr(), getiaddr()	: ... + dirty_set
----------------------------------------------------------------------*/
#define getnode(tf,idx) ((TNODE *)dc_add_r((tf)->cache,(idx)+DC_PAGESIZE))
#define getfnode(tf,idx) ((FREE *)dc_add_r((tf)->cache,(idx)+DC_PAGESIZE))
#define getinode(tf,idx) ((INFO *)dc_add_r((tf)->cache,(idx)+DC_PAGESIZE))

/* Ʒ define ־ address  Ǵ  Ѵ. */
#define getnaddr(tf,idx) ((TNODE *)dc_add_w((tf)->cache,(idx)+DC_PAGESIZE))
#define getfaddr(tf,idx) ((FREE *)dc_add_w((tf)->cache,(idx)+DC_PAGESIZE))
#define getiaddr(tf,idx) ((INFO *)dc_add_w((tf)->cache,(idx)+DC_PAGESIZE))

/* ־ address  dirty_set Ѵ. */
#define reflect_change(tf,p) dc_dirty_set((tf)->cache,(void *)p)

/* wjlee #define putnode(tf,idx,node) (*getnaddr(tf,idx) = *(TNODE *)(node)) */

void putnode(TRIE *tf, IDX idx, TNODE *node)
{
    *(TNODE *)dc_add_w(tf->cache,idx+DC_PAGESIZE)=*node;
}
#define putfnode(tf,idx,node) putnode(tf,idx,(TNODE *)node)
#define putinode(tf,idx,node) putnode(tf,idx,(TNODE *)node)


/*----------------------------------------
  <section> free node management 

  mergefreetree(), splitfreetree()
  :hidden++

  node_alloc(), node_free()
  :hidden
----------------------------------------------------------------------*/
/* 
   mergefreetree()
   ΰ free node tree ϳ ϱ Ͽ ȴ.
   size ū 尡 ׻  ְ ȴ.
   node_alloc() node_free() ȴ.
*/

IDX mergefreetree(TRIE *tf, IDX lidx, IDX ridx)
{
	FREE lfn, rfn;

#ifdef DEBUG_NODE_ALLOC
fprintf(stderr,"m");
#endif
	if(lidx == 0)
		return ridx;
	if(ridx == 0)
		return lidx;

	lfn = *getfnode(tf,lidx);
	rfn = *getfnode(tf,ridx);
	
	/* rfn.size==0̸ ̰ free root̴. */
	/* lfn.size > rfn.size : 񱳿 >=   ,
	  	 >   °, ° . */
	if(rfn.size != 0 && lfn.size > rfn.size)
	{
		lfn.rchild = mergefreetree(tf,lfn.rchild,ridx);
		putfnode(tf,lidx,&lfn);
		return lidx;
	}
	else
	{
		rfn.lchild = mergefreetree(tf,lidx,rfn.lchild);
		putfnode(tf,ridx,&rfn);
		return ridx;
	}
}

/* 
   splitfreetree()
   ϳ free node tree ѷ    Ѵ.
     ġ(address)  key̴.
   node_free() ȴ.
*/
int splitfreetree(TRIE *tf, IDX idx, IDX key, IDX *x, IDX *y)
{
	FREE fn;
#ifdef DEBUG_NODE_ALLOC
fprintf(stderr,"s");
#endif

	if(key == 0)
		return 0;
	if(idx == 0)
	{
	    *x=0;*y=0;
	    return 1;
	}

	fn = *getfnode(tf,idx);
	
	if(key < idx)
	{
	    *y=idx;
	    splitfreetree(tf,fn.lchild,key,x,&(fn.lchild));
	    putfnode(tf,idx,&fn);
	    return 1;
	}
	else /* if(key > idx) */
	{
	    *x=idx;
	    splitfreetree(tf,fn.rchild,key,&(fn.rchild),y);
	    putfnode(tf,idx,&fn);
	    return 1;
	}
}

#define F_DIR_L 0
#define F_DIR_R 1
/*----------------------------------------------------------------------
node_alloc()
	trie node manager
	
----------------------------------------------------------------------*/
IDX node_alloc(TRIE *tf, IDX size)
{
	IDX	idx,pidx,tidx,lidx,ridx;
	FREE	*fnode; /* current free node */
	int 	fdir;

	if(size <= 0)
	{
		fprintf(stderr,"node_alloc:size 0\n");
		exit(0);
	}

#ifdef DEBUG_NODE_ALLOC
fprintf(stderr,"\na");
#endif
	fdir = F_DIR_R;
	pidx = FREE_NODE;
	idx = getfnode(tf,pidx)->rchild;

	while(1)
	{
		fnode = getfnode(tf,idx);

		/*  size 尡 parent:child  
		   parent 忡 ҴѴ.
		   , 1->1->1-> ...̷ Ǿ  տ allocѴ. */
		if(fnode->size == size) /* ũ ġ */
		{
			tidx = mergefreetree(tf,fnode->lchild,fnode->rchild);
			break;
		}
		else
		{
			/* ʿ 켱 Ҵϴ å */
			if(fnode->lchild && 
				getfnode(tf,fnode->lchild)->size >= size)
			{
				pidx = idx;
				idx = fnode->lchild;
				fdir = F_DIR_L;
			}
			else if(fnode->rchild && 
				getfnode(tf,fnode->rchild)->size >= size)
			{
				pidx = idx;
				idx = fnode->rchild;
				fdir = F_DIR_R;
			}
			else
			{
			    /* tidx  pidx child Ű  ȴ.*/
			    /* idx allocate ġ Ÿ. */

				if(fnode->size == 0) /* Ѵ ũ */
				{
					/* տ Ҵ */
					tidx = idx+size;
					putfnode(tf,tidx,fnode);
				}
				else /* if(fnode.size > size)  */
				{ 
			    		/* ڿ Ҵ */
					lidx = fnode->lchild;
					ridx = fnode->rchild;

					fnode->lchild = 0;
					fnode->rchild = 0;
					fnode->size -= size;
					putfnode(tf,idx,fnode);

					tidx = idx;
					idx += fnode->size;
							
					tidx = mergefreetree(tf,tidx,ridx);
					tidx = mergefreetree(tf,lidx,tidx);

				}
				break;
			}
		}
	}
	if(fdir == F_DIR_R)
		getfaddr(tf,pidx)->rchild = tidx;
	else
		getfaddr(tf,pidx)->lchild = tidx;
#ifdef DEBUG_NODE_ALLOC
fprintf(stderr,":%d:%d",idx,size);
#endif
	return idx;
}


/* ----------------------------------------------------------------------
node_free()
	moves nodes to free nodes space 
	and paste free nodes.
----------------------------------------------------------------------*/
void node_free(TRIE *tf, IDX free_idx, IDX size)
{
	IDX	idx,pidx;
	FREE	*fnode,tnode;
	int 	fdir;

 	if(free_idx < 2 || size  <= 0)
	{
		fprintf(stderr,"node_free:idx=%d size=%d\n",free_idx,size);
		exit(0);
	}
#ifdef DEBUG_NODE_ALLOC
fprintf(stderr,"\nf:%d",free_idx);
#endif

	fdir = F_DIR_R;
	pidx = FREE_NODE;
	idx = getfnode(tf,pidx)->rchild;

	/* ϴ Ǵ free node  Ѵ. */
	while(idx)
	{
		fnode = getfnode(tf,idx);
		if(free_idx > idx)
		{
			if(idx + fnode->size == free_idx)
			{
				free_idx = idx;
				if(size != 0)
					size += fnode->size;
				idx = mergefreetree(tf,fnode->rchild,fnode->lchild);
				if(fdir == F_DIR_R)
					getfaddr(tf,pidx)->rchild = idx;
				else
					getfaddr(tf,pidx)->lchild = idx;
			}
			else
			{
				pidx = idx;
				idx = fnode->rchild;
				fdir = F_DIR_R;
			}
		}
		else /* free_idx < idx */
		{
			if(free_idx + size == idx)
			{
				if(fnode->size == 0)
					size = 0;
				else
					size += fnode->size;
				idx = mergefreetree(tf,fnode->rchild,fnode->lchild);
				if(fdir == F_DIR_R)
					getfaddr(tf,pidx)->rchild = idx;
				else
					getfaddr(tf,pidx)->lchild = idx;
			}
			else
			{
				pidx = idx;
				idx = fnode->lchild;
				fdir = F_DIR_L;
			}
		}
	}
	tnode.size = size;
	tnode.rchild = 0;
	tnode.lchild = 0;

	fdir = F_DIR_R;
	pidx = FREE_NODE;
	idx = getfnode(tf,pidx)->rchild;


	/*  ġ free_idx  ִ´. */
	while(idx)
	{
		fnode = getfnode(tf,idx);
		/* fnode->size >= size ? Ȥ >  
		   >=  ...  ׷? */
		if(size != 0 && (fnode->size == 0 || fnode->size >= size))
		{
			/* ũⰡ  Ʒ . */
			pidx = idx;
			if(free_idx < idx)
			{
				idx = fnode->lchild;
				fdir = F_DIR_L;
			}
			else
			{
				idx = fnode->rchild;
				fdir = F_DIR_R;
			}
		}
		else
		{
			/* ڱ ġ  child 籸 */
			splitfreetree(tf,idx,free_idx,&(tnode.lchild),&(tnode.rchild));
			break;
		}
	}

	putfnode(tf,free_idx,&tnode);
	if(fdir == F_DIR_R)
		getfaddr(tf,pidx)->rchild = free_idx;
	else
		getfaddr(tf,pidx)->lchild = free_idx;
}

/*----------------------------------------
  <section> trie basic functions
  : hidden, user tdbm_ Ȥ mdic_ Լ ̿ 
  
  trie_open(), trie_close()
  buf_n_cpy()
  node_search(), info_search()
  trie_insert()
  trie_remove_path()
  trie_search()
----------------------------------------------------------------------*/
TRIE * trie_open(char *fname, int flag, int itype)
{
	TRIE *tf;

	tf = (TRIE *)malloc(sizeof(TRIE));
	tf->info_type = itype;
	tf->cache = dcache_open(fname, flag, sizeof(TNODE));

	if(tf->cache->fend==0) /* new file */
	{
	    /* initialize header & starting information */
		TNODE *tn;
		FREE *fn;
		char *p;

		p=(char *)dc_add_w(tf->cache,0);
		memcpy(p,TRIE_VERSION, TRIE_HEAD_SYSTEM);

		fn=getfaddr(tf,FREE_NODE);
		fn->rchild = 2;		/* 2 is the first free node */

		tn=getnaddr(tf,START_NODE);
		tn->child_size = 0;	/* child_size == 0 means no child */
		tn->child_idx = 0;	/* child_idx ==0, no child */

		fn=getfaddr(tf,2); 	/* real free node */
		fn->size = 0;		/* size == 0 , infinitive size */
		fn->lchild = 0;		/* 0 means no child*/
		fn->rchild = 0;		/* 0 means no child */
		
		tf->state=TRIE_STATE_NEW;
	}
	else /*   */
		tf->state=TRIE_STATE_OLD;

	tf->search_end = 0;
	return tf;
}

/*
	trie_close() trie ϰ ݴ´.
	PUBLIC
*/

void trie_close(TRIE *tf)
{
#ifndef _MSC_VER
	IDX file_size;

	file_size=(getfnode(tf,FREE_NODE)->rchild + 1 + DC_PAGESIZE) * sizeof(TNODE);
	dcache_truncate(tf->cache,file_size);
#endif
	dcache_close(tf->cache);
	free(tf);
}

/* 
   low leve node copy 
   :hidden
*/
void buf_n_cpy(TRIE *tf, IDX des, IDX src, IDX n)
{
	while(n-->0)
		putnode(tf,des++,getnode(tf,src++));
}

/* 
   trie node  siblingȿ N_KEY k ִ ˻ϴ Լ
   :hidden
*/
TNODE *node_search(TRIE *tf, N_KEY k, IDX idx, IDX size, IDX *found)
{
	long l,m,u;	/*  ݵ long ؾ */
	TNODE *node;

	l=0;
	u=size-1;
	while(l<=u) 
	{
		m = (l+u)/2;
		
		node = getnode(tf,idx+m);
		if(k  > node->key)  
			l = m+1;
		else if(k == node->key)  
		{
			*found = m;
			return node;
		}
		else /* if(k  < node->key)  */
			u = m-1;
	}
	*found = l;/* key (ϳ ) ε  */
	return NULL;
}

int inpcmp(INFO *ip1,INFO *ip2)
{
	int i,j;
	char *p,*q;
	j=sizeof(INFO);
	p = (char *)ip1; q = (char *)ip2;
	for(i=0;i<j;i++,p++,q++)
		if(*p>*q) return 1;
		else if(*p<*q) return -1;
	return 0;
}
/* 
   info node  siblingȿ *ip ִ ˻ϴ Լ
   :hidden
*/
INFO *info_search(TRIE *tf, INFO *ip, IDX idx, IDX size, IDX *found)
{
	long l,m,u;	/*  ݵ long ؾ */
	INFO *inode;
	int re;

	l=0;
	u=size-1;
	while(l<=u) 
	{
		m = (l+u)/2;
		
		inode = getinode(tf,idx+m);
		re = inpcmp(ip,inode);
		if(re > 0) 
			l = m+1;
		else if(re == 0)  
		{
			*found = m;
			return inode;
		}
		else /* if(re  < 0)  */
			u = m-1;
	}
	*found = l;/* ip (ϳ ) ε  */
	return NULL;
}

/*
   trie node , str شϴ path node   Լ 
   : hidden
*/
IDX trie_insert(TRIE *tf, N_KEY *str)
{
	TNODE *pn; 		/* previous node*/
	IDX idx, pidx;
	G_SIZE size;
	int i;

	if(str == NULL|| *str == 0)
		return 0;

	for(i=0;i<tf->search_end && tf->search_key[i] == str[i];i++)
		;
	if(i==0)
	{
		tf->search_end = 0;
		tf->search_idx[0] = START_NODE;
		pidx = START_NODE;
	}
	else
	{
		pidx = tf->search_idx[i];
		tf->search_end = i;
		str += i;
	}

	while(*str)
	{
		pn = getnode(tf,pidx);
		idx = pn->child_idx;
		size = pn->child_size;
		if(size == 0)			/* child  */
		{
			TNODE tn={0,0,0,0,0,0};
			idx = node_alloc(tf,1);
			tn.key = *str;
			putnode(tf,idx,&tn);
			pn=getnaddr(tf,pidx);
			pn->child_size = 1;
			pn->child_idx = idx;
		}
		else
		{

			IDX i_pos;
			if(node_search(tf,*str,idx,size,&i_pos) == NULL)
			{
				IDX newidx;
				TNODE tn={0,0,0,0,0,0};

				newidx = node_alloc(tf,size+1);

				pn=getnaddr(tf,pidx);
				pn->child_size++;
				pn->child_idx = newidx;

				buf_n_cpy(tf,newidx,idx,i_pos);
				tn.key = *str;
				putnode(tf,newidx+i_pos,&tn);
				buf_n_cpy(tf,newidx+i_pos+1,idx+i_pos,size-i_pos);

				node_free(tf,idx,size);
				idx = newidx+i_pos;
			}
			else
				idx = idx+i_pos;
		}
		tf->search_key[tf->search_end++] = *str++;
		tf->search_idx[tf->search_end] = idx;
		pidx = idx;
	}
	return pidx;
}
/* 
   :hidden
	trie_search() θ ̴ Լ Դϴ.
	return   ִ.
*/

int trie_search(TRIE *tf, N_KEY *str)
{
	TNODE *node;
	IDX idx,tidx;
	G_SIZE size;
	int i;

	if(str == NULL|| *str == 0)
		return -1;

	for(i=0;i<tf->search_end && tf->search_key[i] == str[i];i++)
		;
	if(i==0)
	{
		tf->search_end = 0;
		tf->search_idx[0] = START_NODE;
		node = getnode(tf,START_NODE);
	}
	else
	{
		node = getnode(tf,tf->search_idx[i]);
		tf->search_end = i;
		str += i;
	}

	while(*str)
	{
		size = node->child_size;
		if(size == 0)
			return 0;
		else
		{
			idx = node->child_idx;
			node = node_search(tf,*str,idx,size,&tidx);
			if(node == NULL)
				return 0;
			idx += tidx;
			tf->search_key[tf->search_end++] = *str++;
			tf->search_idx[tf->search_end] = idx;
			node = getnode(tf,idx);
		}
	}
/*
prn_tnode(node);
*/
	if(node->info_size == 0)
		return 0;
	else
		return tf->search_end;
}


/*
   : hidden
	path node .... info 
*/
int trie_remove_path(TRIE *tf)
{
	TNODE *node;
	TNODE pnode;
	IDX idx, pidx, d_pos,size, tidx, newidx;

	while(tf->search_end)
	{
		idx=tf->search_idx[tf->search_end];
		node=getnode(tf,idx);
		if(node->child_size != 0 || node->info_size != 0)
		    break;

		/* child  data  쿡  .*/
		pidx=tf->search_idx[tf->search_end-1];
		pnode = *getnode(tf,pidx);
		if(pnode.child_size == 1)
		{
			node_free(tf,pnode.child_idx,1);
			pnode.child_size = 0;
			pnode.child_idx = 0;
		}
		else
		{
			tidx = pnode.child_idx;
			d_pos = idx - tidx;
			size = pnode.child_size;

			newidx = node_alloc(tf,size-1);
			buf_n_cpy(tf,newidx,tidx,d_pos);
			buf_n_cpy(tf,newidx+d_pos,idx+1,size-d_pos-1);
			node_free(tf,tidx,size);
			pnode.child_size--;
			pnode.child_idx = newidx;
		}
		putnode(tf,pidx,&pnode);
		tf->search_end--;
	}
	return 1;
}


/*----------------------------------------
  <section> trie basic interface
  :public, trie ⺻ ϴ Լ  public Լ
  
    void trie_write_user_head(TRIE *tf, char *thead);
    void trie_write_user_head(TRIE *tf, char *thead);
    int trie_firstkey(TRIE *tf, N_KEY *str); 
    int trie_lastkey(TRIE *tf, N_KEY *str);  
    int trie_nextkey(TRIE *tf, N_KEY *str); 
    int trie_prevkey(TRIE *tf, N_KEY *str); 
    TRIE *trie_pack(TRIE *tf); 
----------------------------------------------------------------------*/
void trie_write_user_head(TRIE *tf, char *thead)
{
    char *p;
    p=(char *)dc_add_w(tf->cache,TRIE_HEAD_SYSTEM/sizeof(TNODE));
    memcpy(p,thead,TRIE_HEAD_USER);
}

void trie_read_user_head(TRIE *tf, char *thead)
{
    char *p;
    p=(char *)dc_add_r(tf->cache,TRIE_HEAD_SYSTEM/sizeof(TNODE));
    memcpy(thead,p,TRIE_HEAD_USER);

}
int trie_firstkey(TRIE *tf, N_KEY *str)
{
	TNODE *node;
	IDX idx;
	int i=0;

	tf->search_end = 0;
	tf->search_idx[0] = START_NODE;
	node = getnode(tf,START_NODE);

	while(node->child_size)
	{
		idx = node->child_idx;
		node=getnode(tf,idx);
		str[i++] = tf->search_key[tf->search_end++] = node->key;
		tf->search_idx[tf->search_end] = idx;
		if(node->info_size)
			break;
	}
	str[i] = 0;
	return i ;

}
int trie_nextkey(TRIE *tf, N_KEY *str)
{
	TNODE *node,*pnode;
	IDX idx,pidx;
	int i;
	int children, younger_brothers;

	if(tf->search_end > 0)
	{
		idx = tf->search_idx[tf->search_end];
		node = getnode(tf,idx);
		children = node->child_size;

		pidx = tf->search_idx[tf->search_end-1];
		pnode = getnode(tf,pidx);
		younger_brothers = pnode->child_idx + pnode->child_size - idx - 1;
	}

	while(tf->search_end > 0)
	{

		if(children)
		{
			/* DOWN */
			idx = node->child_idx;
			node=getnode(tf,idx);
			tf->search_key[tf->search_end++] = node->key;
			tf->search_idx[tf->search_end] = idx;
			if(node->info_size > 0)
			    	break;
			younger_brothers = children-1;
			children = node->child_size;
		}
		else if(younger_brothers) /* no children && younger_brothers */
		{
			/* RIGHT */
			idx++;
			node = getnode(tf,idx);
			tf->search_key[tf->search_end-1] = node->key;
			tf->search_idx[tf->search_end] = idx;
			if(node->info_size > 0)
			    	break;
			younger_brothers--;
			children = node->child_size;
		}
		else /* no children && no younger_brothers */
		{
			/* UP */
			tf->search_end--;
			if(tf->search_end == 0)
				break;

			idx = tf->search_idx[tf->search_end];
			children = 0;
			pidx = tf->search_idx[tf->search_end-1];
			pnode = getnode(tf,pidx);
			younger_brothers = pnode->child_idx+pnode->child_size - idx - 1;
		}
	}
	for(i=0;i<tf->search_end;i++)
		str[i] = tf->search_key[i];
	str[i] = 0;
	return i ;
}
int trie_lastkey(TRIE *tf, N_KEY *str)
{
	TNODE *node;
	IDX idx;
	int i=0;

	tf->search_end = 0;
	tf->search_idx[0] = START_NODE;
	node = getnode(tf,START_NODE);

	while(node->child_size)
	{
		idx = node->child_idx + node->child_size - 1;
		node=getnode(tf,idx);
		str[i++] = tf->search_key[tf->search_end++] = node->key;
		tf->search_idx[tf->search_end] = idx;
	}
	str[i] = 0;
	return i ;
}

int trie_prevkey(TRIE *tf, N_KEY *str)
{
	TNODE *node;
	IDX idx,pidx;
	int i;

	idx = tf->search_idx[tf->search_end];

	while(tf->search_end > 0)
	{
		pidx = tf->search_idx[tf->search_end-1];
		if(getnode(tf,pidx)->child_idx < idx) /* elder_brothers */
		{
			/* LEFT */
			idx--;
			node = getnode(tf,idx);
			tf->search_key[tf->search_end-1] = node->key;
			tf->search_idx[tf->search_end] = idx;

			/* DOWN */
			while(node->child_size)
			{
			    idx = node->child_idx+node->child_size-1;
			    node=getnode(tf,idx);
			    tf->search_key[tf->search_end++] = node->key;
			    tf->search_idx[tf->search_end] = idx;
			}
			if(node->info_size > 0)
				break;
		}
		else /* parent */
		{
			/* UP */
			tf->search_end--;
			if(tf->search_end == 0)
				break;

			idx = tf->search_idx[tf->search_end];
			node = getnode(tf,idx);
			if(node->info_size > 0)
				break;
		}
	}
	for(i=0;i<tf->search_end;i++)
		str[i] = tf->search_key[i];
	str[i] = 0;
	return i ;
}

/* trie_pack   ־   . */
IDX tnode_pack(TRIE *tf, TRIE *tt, G_SIZE size, IDX idx)
{
	IDX tidx;
	int i,j;
	TNODE tnode, *node;
	INFO *inodep;

	if(size == 0)
		return 0;

	tidx = node_alloc(tt,size);
	for(i=0;i<size;i++)
	{
		node = getnode(tf,idx+i);
		tnode.key = node->key;
		tnode.info_size = node->info_size;
		tnode.child_size = node->child_size;
		tnode.child_idx=node->child_idx;
		if(tnode.info_size > 0)
		{
			tnode.info_idx = node_alloc(tt,tnode.info_size);
			for(j=0;j<tnode.info_size;j++)
			{
				inodep = getinode(tf,node->info_idx+j);
				putinode(tt,tnode.info_idx+j,inodep);
			}
		}
		else
			tnode.info_idx = 0;
		tnode.child_idx = tnode_pack(tf,tt,tnode.child_size, tnode.child_idx);
		putnode(tt,tidx+i,&tnode);
	}
	return tidx;
}

#define TRIE_FILE_SUFFIX ".tri"
#define OLD_FILE_SUFFIX ".old"
#define TMP_FILE_SUFFIX ".tmp"

TRIE *trie_pack(TRIE *tf)
{
	char newfilename[256];
	char oldfilename[256];
	char triefilename[256];
	char headbuff[TRIE_HEAD_SIZE];
	TRIE *tt;
	TNODE tnode,*node;
	int oldflag;
	int olditype;

	oldflag=tf->cache->flag;
	olditype=tf->info_type;

	strcpy(triefilename,tf->cache->fname);
	strcpy(newfilename,tf->cache->fname);
	strcat(newfilename,TMP_FILE_SUFFIX);
	strcpy(oldfilename,tf->cache->fname);
	strcat(oldfilename,OLD_FILE_SUFFIX);

	unlink(newfilename);
	tt = trie_open(newfilename, FILE_FLAG_W, olditype);
	trie_read_user_head(tf,headbuff);
	trie_write_user_head(tt,headbuff);

	node = getnode(tf,START_NODE);
	tnode.child_size = node->child_size;
	tnode.child_idx = tnode_pack(tf,tt,node->child_size,node->child_idx);
	putnode(tt,START_NODE,&tnode);

	fprintf(stderr,"trie packed %s!!\n",newfilename);
	fprintf(stderr,"closing %s\n",newfilename);
	fprintf(stderr,"closing %s\n",tf->cache->fname);
	trie_close(tt);
	trie_close(tf);
	unlink(oldfilename);
	rename(triefilename,oldfilename);
	fprintf(stderr,"rename %s --> %s\n",triefilename,oldfilename);
	rename(newfilename,triefilename);
	fprintf(stderr,"rename %s --> %s\n",newfilename,triefilename);
	fprintf(stderr,"opening new trie %s",triefilename);
	return trie_open(triefilename, oldflag, olditype);
}

/*----------------------------------------
  <section> application I -- MDIC
  : public

  ¼ м⸦  , INFO 带  ȰѴ.
----------------------------------------------------------------------*/
#ifdef DEBUG_LOOK
long wjdsearch=0;
#endif

#define MDIC_FILE_SUFFIX ".mdc"

MDIC * mdic_open(char *fname, char type)
{
	int flags;
	char filename[256];

	if(type == 'r')
		flags = FILE_FLAG_R ;
	else if(type == 'w')
		flags = FILE_FLAG_W ;
	else
	{
		fprintf(stderr,"mdic_open:open_type unknown '%c'\n",type);
		return NULL;
	}

	strcpy(filename,fname);
	strcat(filename,MDIC_FILE_SUFFIX);
	return (MDIC *)trie_open(filename,flags,ITYPE_MDIC);
}
void mdic_close(MDIC *tf)
{
#ifdef DEBUG_LOOK
	fprintf(stderr,"mdic lookup %d\n",wjdsearch);
#endif

	trie_close((TRIE *)tf);
}

/* 
    ܾ Է
*/

int mdic_store(MDIC *tf, N_KEY *word, INFO *inode)
{
	TNODE node; /* previous node and current node */
	INFO *i_node;
	IDX idx, info;

	idx=trie_insert(tf,word);
	if(idx==0)
		return 0;

	node = *getnode(tf,idx);

	if(node.info_size == 0)
	{
		info = node_alloc(tf,1); /* wjlee bug fix . size+1 --> 1 */
		node.info_idx = info;
		node.info_size = 1;
		putnode(tf,idx,&node);
		putinode(tf,info,inode);
	}
	else
	{
		IDX i_pos;

		i_node = info_search(tf,inode,node.info_idx,node.info_size,&i_pos);
		if(i_node == NULL)
		{
			info = node_alloc(tf,node.info_size+1);
			buf_n_cpy(tf,info,node.info_idx,i_pos);

			putinode(tf,info+i_pos,inode);
			buf_n_cpy(tf,info+i_pos+1,node.info_idx+i_pos,node.info_size-i_pos);
			node_free(tf,node.info_idx,node.info_size);
			node.info_size++;
			node.info_idx = info;
			putnode(tf,idx,&node);
		}
		else
		{
			fprintf(stderr,"%s:warning:'%s' infomation duplicated!\n",__FILE__,word);
		}
	}
	return 0;
}

/*
	look up substring
	PUBLIC
*/
int mdic_lookup(MDIC *tf, N_KEY *str, TNODE tpath[], int tp_size)
{
	int i;

#ifdef DEBUG_LOOK
	wjdsearch++;
#endif

	if(trie_search(tf,str) < 0)
		return -1; 		/*  ã  */
	if(tp_size < tf->search_end -1)
		return -2; 		/* tp_size Ű Ѵ */
	else
	{
		for(i=0;i<tf->search_end;i++)
			tpath[i] = *getnode(tf,tf->search_idx[i+1]);
		return i; /* path length 0..n */
	}
}

/*
   new 990830
    徿 ϱ Ͽ
*/

/* start from idx = 1*/
IDX mdic_node_look(MDIC *tf, N_KEY key, IDX idx)
{
	IDX t;
	TNODE *node;
	node = getnode(tf,idx);

	if(node->child_size != 0 && 
		node_search(tf,key,node->child_idx,node->child_size,&t) != NULL)
		    return node->child_idx+t;
	return 0;
}

/*
exact match
return value means
	-1 : error input
	0 : not exist
	other : the idx of TNODE that includes info_idx
	PUBLIC
*/
int mdic_fetch(MDIC *tf, N_KEY *str, INFO *inode)
{
	TNODE *node;
	IDX idx;
	IDX isize, iidx;
	IDX i;
	int x;

	x = trie_search(tf,str);
	if(x <= 0)
		return x;
	else
	{
	    	idx=tf->search_idx[x];
		node = getnode(tf,idx);
		isize=node->info_size;
		iidx=node->info_idx;
		for(i=0;i<isize;i++)
			inode[i] = *getinode(tf,iidx+i);
		return isize;
	}
}
/*
	PUBLIC
*/
int mdic_delete(MDIC *tf, N_KEY *str, INFO *ip)
{
	TNODE node;
	IDX idx;
	int x;

	x = trie_search(tf,str); /* ϴ ã .*/

	if(x <= 0 ) /*  κ  .... delete ־ */
		return x;

	idx=tf->search_idx[x];
	node = *getnode(tf,idx);
	if(node.info_size != 0)
	{
		IDX i_pos;
		IDX info;
		INFO *i_node;

		i_node = info_search(tf,ip,node.info_idx,node.info_size,&i_pos);
		if(i_node != NULL)
		{
			if(node.info_size > 1)
			{
				info = node_alloc(tf,node.info_size-1);
				buf_n_cpy(tf,info,node.info_idx,i_pos);
				buf_n_cpy(tf,info+i_pos,node.info_idx+i_pos+1,node.info_size-i_pos-1);
			}
			else
				info=0;
			node_free(tf,node.info_idx,node.info_size);
			node.info_size--;
			node.info_idx = info;
			putnode(tf,idx,&node);
		}
		else
			return 0;
	}
	trie_remove_path(tf);
	return 1;
}

/*----------------------------------------
  <section> application II -- TDBM
  : public

    TDBM : trie based dbm like file system
     κ dbm ܾ 10 Ѿ ӵ ʹ 
    (5 30, 25 50), trie ̿Ͽ dbm
    ϰ .
     ӵ 25 ܾ 5 ɸ.
     ̸   (parameter) ణ ̰ ִ.
----------------------------------------------------------------------*/

#define TDBM_FILE_SUFFIX ".tdb"

TDBM * tdbm_open(char *fname, char type)
{
	int flags;
	char filename[256];

	if(type == 'r')
		flags = FILE_FLAG_R ;
	else if(type == 'w')
		flags = FILE_FLAG_W ;
	else
	{
		fprintf(stderr,"tdbm_open:open_type unknown '%c'\n",type);
		return NULL;
	}

	strcpy(filename,fname);
	strcat(filename,TDBM_FILE_SUFFIX);
	return (TDBM *)trie_open(filename,flags,ITYPE_TDBM);
}


void tdbm_close(TDBM *tf)
{
    trie_close((TRIE *)tf);
}

int tdbm_store(TDBM *tf, N_KEY *word, N_KEY *data, int store_mode)
{
	TNODE node; /* previous node and current node */
	IDX idx;
	INFO *inode;
	info_head head; 	/* data size ũϱ  head д.*/
	IDX i;

	idx=trie_insert(tf,word);
	if(idx==0)
		return -1;

	node = *getnode(tf,idx);

	if(node.info_size != 0)
		if(store_mode == TDBM_INSERT)
			return 1; /* exist */
		else
		{
			head=*(info_head*)getnode(tf,node.info_idx);
			node_free(tf,head.info,head.size);
		}
	else
	{
		node.info_idx=node_alloc(tf,1);
		node.info_size=1;
		putnode(tf,idx,&node);
	}

	head.size=strlen(data)/12+1; /* sizeof INFO +1 */
	head.info=node_alloc(tf,head.size);
	putinode(tf,node.info_idx,&head);

	inode=(INFO *)data;
	for(i=0;i<head.size;i++,inode++)
		putinode(tf,head.info+i,inode);

	return 0;
}

int tdbm_fetch(TDBM *tf, N_KEY *str, N_KEY *data)
{
	TNODE *node;
	INFO *inode;
	info_head head;
	IDX idx;
	IDX i;
	int x;

	inode=(INFO *)data;
	data[0]=0;

	x = trie_search(tf,str);
	if(x <= 0)
		return x;
	else
	{
		idx=tf->search_idx[x];
		node = getnode(tf,idx);
		if(node->info_size)
		    head = *(info_head *)getnode(tf,node->info_idx);
		for(i=0;i<head.size;i++)
			inode[i] = *getinode(tf,head.info+i);
		return head.size;
	}
}

int tdbm_delete(TDBM *tf, N_KEY *str)
{
	TNODE node;
	info_head head;
	IDX idx;
	int x;

	x = trie_search(tf,str); /* ϴ ã .*/

	if(x <= 0 ) 
		return x;

	idx=tf->search_idx[x];
	node = *getnode(tf,idx);

	/* trie_search ߴٸ node.info_size 1 ̴̻. */
	head=*(info_head*)getnode(tf,node.info_idx);
	node_free(tf,head.info,head.size);
	node_free(tf,node.info_idx,1);
	node.info_size = 0;
	node.info_idx = 0;
	putnode(tf,idx,&node);
	trie_remove_path(tf);
	return 1;
}


/*----------------------------------------
  <section> application III -- TX
  : public

    TX : trie based indexing system
    ܾ : ġ(Ϲȣ, ϳ ġ), ġ, ġ ....
    ̷   ̴.
    ġ     ϳ ġ    ũ
    12Ʈ̴.
    firstkery, lastkery, nextkery trie_firstkey  ϸȴ.
----------------------------------------------------------------------*/

#define TX_FILE_SUFFIX ".tx"

TX * tx_open(char *fname, char type)
{
	int flags;
	char filename[256];

	if(type == 'r')
		flags = FILE_FLAG_R ;
	else if(type == 'w')
		flags = FILE_FLAG_W ;
	else
	{
		fprintf(stderr,"tx_open:open_type unknown '%c'\n",type);
		return NULL;
	}

	strcpy(filename,fname);
	strcat(filename,TX_FILE_SUFFIX);
	return (TX *)trie_open(filename,flags,ITYPE_TX);
}


void tx_close(TX *tf)
{
    trie_close((TRIE *)tf);
}

int tx_store(TX *tf, N_KEY *word, txtype *p)
{
	TNODE node; /* previous node and current node */
	txtype *txinfo;
	IDX idx, info;
	IDX newidx;
	int newword;

	idx=trie_insert(tf,word);
	if(idx == 0)
		return -1;

	node = *getnode(tf,idx);
	info = node.info_idx;

	if(node.info_size != 0)
	{
		newidx = node_alloc(tf,1);
		txinfo=(txtype *)getiaddr(tf,info);
		p->next=txinfo->next;
		txinfo->next=newidx;
		txinfo->posx++;
		putnode(tf,newidx,(TNODE *)p);
		newword=0;
	}
	else
	{
		newidx = node_alloc(tf,2);
		putnode(tf,newidx+1,(TNODE *)p);
		p->next=0;
		txinfo=(txtype *)getiaddr(tf,newidx);
		txinfo->posx=1;
		txinfo->next=newidx+1;
		node.info_idx=newidx;
		node.info_size=1;
		putnode(tf,idx,&node);
		newword=1;
	}
	return newword;
}

int tx_fetch(TX *tf, N_KEY *str, txtype *p)
{
	TNODE *node;
	txtype *txinfo;
	IDX idx,next;
	IDX size;
	IDX i;
	int x;

	x = trie_search(tf,str);

	if(x <= 0)
		return x;
	else
	{
		idx=tf->search_idx[x];
		node = getnode(tf,idx);
		txinfo = (txtype *)getnode(tf,node->info_idx);
		size=txinfo->posx;
		next=txinfo->next;
		for(i=0;i<size;i++)
		{
			p[i] = *(txtype *)getnode(tf,next);
			next=p[i].next;
		}
		return size;
	}
}

int tx_count(TX *tf, N_KEY *str)
{
	TNODE *node;
	txtype *txinfo;
	IDX idx;
	int x;

	x = trie_search(tf,str);
	if(x <= 0)
		return x;
	else
	{
		idx=tf->search_idx[x];
		node = getnode(tf,idx);
		txinfo = (txtype *)getnode(tf,node->info_idx);
		return txinfo->posx;
	}
}

int txcmp(txtype *p, txtype *q)
{
	if(p->posx - q->posx)
		return p->posx - q->posx;
	if(p->posy - q->posy)
		return p->posy - q->posy;
	return 0;
}

int tx_delete(TX *tf, N_KEY *str, txtype *p)
{
	IDX i;
	int x;
	TNODE node;
	txtype txinfo,txkill;
	IDX idx,size,next,kill,pidx;

	x = trie_search(tf,str); /* ϴ ã .*/

	if(x <= 0 ) 
		return x;

	idx=tf->search_idx[x];
	node = *getnode(tf,idx);
	if(node.info_size != 0)
	{
		txinfo=*(txtype *)getnode(tf,node.info_idx);
		next=txinfo.next;
		size=txinfo.posx;
		if(p==NULL)
		{
			node_free(tf,node.info_idx,1);
			for(i=0;i<size;i++)
			{
				txkill = *(txtype *)getnode(tf,next);
				node_free(tf,next,1);
				next=txkill.next;
			}
			node.info_size = 0;
			node.info_idx = 0;
			putnode(tf,idx,&node);
		}
		else
		{
			pidx=node.info_idx;
			for(i=0;i<size;i++)
			{
				txkill = *(txtype *)getnode(tf,next);
				if(txcmp(p,&txkill)==0)
					break;
				pidx=next;
				next=txkill.next;
			}
			if(i<size)
			{
				kill=next;
				next=txkill.next;
				node_free(tf,kill,1);
				txinfo.posx--;
				if(txinfo.posx==0)
				{
					node_free(tf,node.info_idx,1);
					node.info_idx=0;
					node.info_size=0;
					putnode(tf,idx,&node);
				}
				else
				{
					txkill = *(txtype *)getnode(tf,pidx);
					txkill.next=next;
					putnode(tf,pidx,(TNODE *)&txkill);
					if(pidx!=node.info_idx)
						putnode(tf,node.info_idx,(TNODE *)&txinfo);
				}
			}
		}
	}
	trie_remove_path(tf);
	return 1;
}

/*----------------------------------------
  <section> application IV -- TY
  :public

----------------------------------------------------------------------*/
#define TY_FILE_SUFFIX ".ty"

TY * ty_open(char *fname, char type)
{
	int flags;
	char filename[256];

	if(type == 'r')
		flags = FILE_FLAG_R ;
	else if(type == 'w')
		flags = FILE_FLAG_W ;
	else
	{
		fprintf(stderr,"ty_open:open_type unknown '%c'\n",type);
		return NULL;
	}

	strcpy(filename,fname);
	strcat(filename,TY_FILE_SUFFIX);
	return (TY *) trie_open(filename,flags,ITYPE_TY);
}


void ty_close(TY *tf)
{
    trie_close((TRIE *)tf);
}

int ty_store(TY *tf, N_KEY *word, char *p, int ysize, int plen)
{
	TNODE node; /* previous node and current node */
	INFO *inode;
	txtype txinfo;
	IDX idx;
	IDX newidx;
	IDX plen_size;
	IDX i;

	idx=trie_insert(tf,word);
	if(idx == 0)
		return -1;

	node = *getnode(tf,idx);

	if(node.info_size != 0)
	{
		txinfo=*(txtype *)getnode(tf,node.info_idx);
		node_free(tf,txinfo.next,txinfo.posx/12+1);
	}
	else
	{
		newidx = node_alloc(tf,1);
		node.info_idx=newidx;
		node.info_size=1;
	}
	plen_size=plen/12+1;

	txinfo.next = node_alloc(tf,plen_size);
	txinfo.posx=plen;
	txinfo.posy=ysize;
	putnode(tf,node.info_idx,(TNODE *)&txinfo);
	putnode(tf,idx,&node);
	inode=(INFO *)p;
	for(i=0;i<plen_size;i++,inode++)
		putinode(tf,txinfo.next+i,inode);
	return 0;
}

int ty_fetch(TY *tf, N_KEY *str, char *p)
{
	TNODE *node;
	INFO *inode;
	txtype *txinfo;
	IDX idx;
	IDX size,next,rsize;
	IDX i;
	int x;

	x = trie_search(tf,str);
	if(x <= 0)
		return x;
	else
	{
		idx=tf->search_idx[x];
		node = getnode(tf,idx);
		txinfo = (txtype *)getnode(tf,node->info_idx);

		size=txinfo->posx/12+1; /* node size */
		next=txinfo->next;
		rsize=txinfo->posy;
		inode=(INFO *)p;
		for(i=0;i<size;i++)
			inode[i] = *getinode(tf,next+i);
		return rsize; /* real data size */
	}
}

int ty_count(TY *tf, N_KEY *str)
{
	TNODE *node;
	txtype *txinfo;
	IDX idx;
	int x;

	x = trie_search(tf,str);
	if(x <= 0)
		return x;
	else
	{
		idx=tf->search_idx[x];
		node = getnode(tf,idx);
		txinfo = (txtype *)getnode(tf,node->info_idx);
		return txinfo->posy;
	}
}

/*----------------------------------------
  <section> debugging & tracing
  : public
----------------------------------------------------------------------*/
void prn_info(INFO *inode)
{
        fprintf(stderr,"INFO\tWord_ID = %ld,\n",inode->Word_ID);
        fprintf(stderr,"\tmtag = %d,\n",inode->mtag);
        fprintf(stderr,"\tstag = %d,\n",inode->stag);
        fprintf(stderr,"\tphone = %d,\n",inode->phone);
        fprintf(stderr,"\tgram = %d,\n",inode->gram);
        fprintf(stderr,"\tsem = %d,\n",inode->sem);
        fprintf(stderr,"\tdomain = %d,\n",inode->domain);
        fprintf(stderr,"\tdummy = %d,\n",inode->dummy);
}
void prn_tnode(TNODE *tnode)
{
        fprintf(stderr,"TNODE\tkey = %ld,\n",tnode->key);
        fprintf(stderr,"\tdummy = %d,\n",tnode->dummy);
        fprintf(stderr,"\tinfo_size = %d,\n",tnode->info_size);
        fprintf(stderr,"\tchild_size = %d,\n",tnode->child_size);
        fprintf(stderr,"\tinfo_idx = %ld,\n",tnode->info_idx);
        fprintf(stderr,"\tchild_idx = %ld,\n",tnode->child_idx);
}
void prn_free(FREE *fnode)
{
        fprintf(stderr,"FREE\tsize = %ld,\n",fnode->size);
        fprintf(stderr,"\tlchild = %ld,\n",fnode->lchild);
        fprintf(stderr,"\trchild = %ld,\n",fnode->rchild);
}


int trace_free(TRIE *tf, IDX idx)
{
    char xline[1024];
    FREE fn;
    int x=0;
    fn=*getfnode(tf,idx);

    fprintf(stderr,"\nidx=%ld\n",idx);
    prn_free(&fn);

    while(1)
    {
	printf("l,r,u,q>");
	fgets(xline,1024,stdin);
	if(strcmp(xline,"q")==0)
	    return 1;
	if(strcmp(xline,"u")==0)
	    break;
	else if(strcmp(xline,"l")==0)
	    x=trace_free(tf,fn.lchild);
	else if(strcmp(xline,"r")==0)
	    x=trace_free(tf,fn.rchild);
	if(x==1)
	    return x;
	prn_free(&fn);
    }
    return 0;
}


int trace_tnode(TRIE *tf, IDX idx)
{
    char xline[1024];
    TNODE tn;
    int x=0;

    if(idx==0)
	return 0;
    tn=*getnode(tf,idx);
    fprintf(stderr,"\nidx=%ld\n",idx);
    prn_tnode(&tn);

    while(1)
    {
	printf("s,c,u,i,q>");
	fgets(xline,1024,stdin);
	if(strcmp(xline,"q")==0)
	    return 1;
	if(strcmp(xline,"u")==0)
	    break;
	else if(strcmp(xline,"s")==0)
	    x=trace_tnode(tf,idx+1);
	else if(strcmp(xline,"c")==0)
	    x=trace_tnode(tf,tn.child_idx);
	else if(strcmp(xline,"i")==0)
	{
	    INFO *inode;
	    inode=getinode(tf,tn.info_idx);
	    prn_info(inode);
	}
	if(x==1)
	    return x;
	prn_tnode(&tn);
    }
    return 0;
}

int count_free(TRIE *tf, IDX idx)
{
    FREE fn;
    int x=1;
    fn=*getfnode(tf,idx);

    if(fn.lchild)
	x+=count_free(tf,fn.lchild);
    if(fn.rchild)
	x+=count_free(tf,fn.rchild);
    return x;
}

int do_trace(TRIE *tf)
{
    char xline[1024];
    int x=0;
    while(1)
    {
	printf("f,t,fc,q>");
	fgets(xline,1024,stdin);
	if(strcmp(xline,"q")==0)
	    break;
	else if(strcmp(xline,"f")==0)
	    trace_free(tf,FREE_NODE);
	else if(strcmp(xline,"t")==0)
	    trace_tnode(tf,START_NODE);
	else if(strcmp(xline,"fc")==0)
	{
	    x=count_free(tf,FREE_NODE);
	    fprintf(stderr,"<fcount> n of fnode =%d\n",x);

	}
    }
}
