#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <iostream.h>
#include <stdiostream.h>

#include <LEDA/string.h>
#include <LEDA/set.h>
#include <LEDA/list.h>
#include <LEDA/graph.h>

#include <LEDA/graph_alg.h>
#include <LEDA/graphwin.h>

#include <trie.h>

#define MAXLINE 2048
#define TMPFILE "morphtmp"
#define PWT_TDBM_FILE "STAT/PWT"
#define PTT_TDBM_FILE "STAT/PTT"

#define EOS "eos"
#define BNK "bnk"

struct NodeInfo {
	string data;
	double prob;
	node backPtr;	// for viterbi

	NodeInfo() : data(""), prob(0.0), backPtr(NULL) {;}
	NodeInfo(const char* str) : data(str), prob(0.0), backPtr(NULL) {}
	~NodeInfo() {}

	NodeInfo& operator=(const NodeInfo& x) { 
		data = x.data;
		prob = x.prob;
		backPtr = x.backPtr;
		return *this;
	}

	friend istream& operator>>(istream& I, NodeInfo& ni)
	{ I >> ni.data >> ni.prob >> ni.backPtr; return I; }
	friend ostream& operator<<(ostream& O, const NodeInfo& ni) 
	{ return O << ni.data << '[' << ni.prob << ']' << '{' << ni.backPtr << '}'; } 
};

TDBM *pwt_tf, *ptt_tf;
GRAPH<NodeInfo,double> hmm;


/* Graph Drawing - START */
void FixNode(GraphWin& gw, node v)
{ // position no used
  if (gw.get_color(v) != red)
     gw.set_color(v,red);
  else
     gw.set_color(v,yellow);
}

void UpdateDrawing(GraphWin& gw, int steps=400)
{ const graph& G = gw.get_graph();
  double xleft   = gw.get_xmin();
  double xright  = gw.get_xmax();
  double ybottom = gw.get_ymin();
  double ytop    = gw.get_ymax();
  node_array<double> xpos(G);
  node_array<double> ypos(G);
  node v;
  forall_nodes(v,G) {
    point p = gw.get_position(v);
    xpos[v] = p.xcoord();
    ypos[v] = p.ycoord();
  }

  //list<node> fixed = gw.get_selected_nodes();

  list<node> fixed;
  forall_nodes(v,G)
      if (gw.get_color(v) == red) fixed.append(v);

  SPRING_EMBEDDING(G,fixed,xpos,ypos,xleft,xright,ybottom,ytop,steps);
  gw.set_position(xpos,ypos);
}

list<string> PerlSplit(string org, const char* delim) 
{
	list<string> ret;
	string* str;
	char* tok =	strtok(org.cstring(), delim);
	if (tok == NULL) { return ret; }

	str = new string(tok);
	ret.append(*str);
	while((tok = strtok(NULL, "+\n")) != NULL) {
		str = new string(tok);
		ret.append(*str);
	}
	return ret;
}

void UpdateProbScore(node from, node to)
{
	static int cnt = 0;
	static unsigned char data[256];
	double PTT, PWT;
	int x;

	//   Ȯ P(T_t|T_i-1)  
	string ttS = hmm[from].data + "|" +  hmm[to].data;	// Bigram tags
	if (tdbm_fetch(ptt_tf, (unsigned char*)ttS.cstring(), data)) 	
		PTT = atof((const char*)data);
	else PTT = 0.01;	// Smoothing Factor

	//  P(W|T)  
	unsigned char* wtS = (unsigned char*) hmm[to].data.cstring();;	
	if (tdbm_fetch(pwt_tf, wtS, data)) 	
		PWT = atof((const char*)data);
	else PWT = 0.01;	// Smoothing Factor
	
	double P = hmm[from].prob * PTT * PWT;
	if (hmm[to].backPtr == NULL || P > hmm[to].prob) {
		hmm[to].backPtr = from;
		hmm[to].prob = P; 
	}
	cout << cnt++ << " " << hmm.inf(from) << "->" << hmm.inf(to) << "\n"; 
}

int main()
{
	// 1.  " Էϼ : " 
	char line[MAXLINE];
	cout << "濬 °Դϴ.\n";
	cout << " Էϼ.\n";
	cout << " >> ";
	gets(line);
	cout << "Էµ : " << line << "\n";

	// 2.  ¼ ؼ⸦ .
	cout << "\n------------- 1. ¼ ؼ⸦ մϴ. --------------\n";
	char cmd[MAXLINE];
	sprintf(cmd,
		"rsh kibs \"echo '%s' | /user3/yuntan/hmorph_new/hmorph\" | TransTagSet.pl> %s",
		line, TMPFILE);
cout << cmd << "\n";
	if ( system(cmd) == -1 ) {
		cerr << "System failed!\n";
		return -1;
	}
	sleep(1);

	// 3.  2  ͼ ׷(DAG) Ѵ.
	NodeInfo ni(EOS);
	node start, end;
	end = hmm.new_node(ni);	//  .
	hmm[end].prob = 1.0;

	cout << "\n------------- 2. ¼ ؼ Դϴ. --------------\n";
	FILE *fp;
	while( (fp = fopen(TMPFILE, "r") ) == NULL)	;
	while( fgets(line, MAXLINE, fp) ) {
		cout << line;
		if (line[1] == '<') //   
		{ 
			NodeInfo ni(strtok(line, "<> "));
			start = end;
			end = hmm.new_node(BNK);
		}
		else if (strlen(line) > 2)	//   ¼ 
		{
			node v, w;
			char* tmp_tok;

			NodeInfo ni(strtok(line, "+\n"));
			v = hmm.new_node(ni);
			hmm.new_edge(start,v);

			while((tmp_tok = strtok(NULL, "+\n")) != NULL) {
				NodeInfo ni(tmp_tok);
				w = hmm.new_node(ni);
				hmm.new_edge(v, w);
				v = w;
			}
			if (v != w) hmm.new_edge(v,end);
			else hmm.new_edge(w,end);
		}
	}
	NodeInfo ni2(EOS);
	hmm[end] = ni2; 

cout << "---- Graph ----";
cout << hmm << "\n";

	// 4. ͺ ˰ Ѵ.
	// 4.  Ȯ Ѵ. (HMM)
	// 	1) Node  - P(Wt|Ti)
	//	2) Edge  - P(Ti|Tj)
	// Ȯ TDBM Ǿ .. ʱȭ
	pwt_tf = tdbm_open(PWT_TDBM_FILE, 'r');
	ptt_tf = tdbm_open(PTT_TDBM_FILE, 'r');

	// Viterbi Algorithm by ALLEN Book
	// 1. Score value setting by breadth first search
	list<node> nodeQ;
	node v, w, fnode = hmm.first_node();
	nodeQ.append(fnode);
	while(!nodeQ.empty()) {
		v = nodeQ.pop();
		edge e;
		forall_out_edges(e, v) { 
			w = target(e);
			if (nodeQ.empty()) nodeQ.append(w);
			else if ( !(w==nodeQ.tail()) ) nodeQ.append(w);
			UpdateProbScore(v,w);
		}
	}
	tdbm_close(pwt_tf);
	tdbm_close(ptt_tf);

	// 2. ݴ 󰡸鼭 
	cout << "------------- RESULT --------------\n";
	v = end;
	while(v) {
		cout << hmm.inf(v) << "\n";
		v = hmm.inf(v).backPtr;
	}

	// 6.  쿡 ش.
	// ׷ ش.
	GraphWin gw(hmm);
	gw.set_node_label_type(data_label);
	gw.set_node_shape(rectangle_node);
	gw.set_edge_direction(directed_edge);
	gw.set_node_width(40);
	gw.set_node_height(40);

	int dx = 10; int dy = 10; int x0 = dx; int y0 = 100;
	node first = hmm.first_node();
	point pos(x0, y0);
	gw.set_position(first, pos);

	FixNode(gw, first);
	UpdateDrawing(gw);
	gw.display(window::center, window::center);
	gw.zoom_graph();
	while(gw.edit())    ;
	
	return 1;
}

