#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "trie.h"
#include "phtag.c"

#define MAXLINE 2048
#define TRUE	1
#define FALSE	0
#define PWT_TDBM_FILE "STAT/PWT"
#define PTT_TDBM_FILE "STAT/PTT"

#define EOS "eos"
#define BNK "bnk"

typedef struct MNode {
	char mout_str[128];
	char wp_tag[5];
	double prob_wt;
	double prob;
	int backptr;	/* for viterbi */
	int sibling;
} MNode;

typedef struct WPhead{
	char word[128];
	int mnode;
} WPhead;

WPhead wp[1024];
int wp_end=1;

MNode mn[2048];
int mn_end=1;

int new_wp(char *str)
{
	strcpy(wp[wp_end].word, str);
	wp[wp_end].mnode = 0;
	return wp_end++;
}

int new_mnode(char *str, char *wp_tag, double prob)
{
	strcpy(mn[mn_end].mout_str,str);
	strcpy(mn[mn_end].wp_tag, wp_tag);
	mn[mn_end].prob_wt=prob;
	mn[mn_end].backptr=0;
	mn[mn_end].sibling=0;
	return mn_end++;
}

TDBM *pwt_tf, *ptt_tf;

void UpdateProbScore(int from, int to)
{
	static int cnt = 0;
	unsigned char data[256];
	unsigned char TTS[256];
	double PTT, PWT;
	double P;
	int x;

	/*   Ȯ P(T_t|T_i-1)  
	*/
	sprintf(TTS,"%s_%s",mn[from].wp_tag,mn[to].wp_tag);
	if (tdbm_fetch(ptt_tf, TTS, data)>0) 	
		PTT = atof((const char*)data);
	else PTT = 0.01;	/* Smoothing Factor */

/*
	//  P(W|T)  
	unsigned char* wtS = (unsigned char*) hmm[to].data.cstring();;	
	if (tdbm_fetch(pwt_tf, wtS, data)) 	
		PWT = atof((const char*)data);
	else PWT = 0.01;	// Smoothing Factor
*/
	
	P = mn[from].prob * PTT * mn[to].prob_wt;
	if (mn[to].backptr == 0 || P > mn[to].prob) {
		mn[to].backptr = from;
		mn[to].prob = P; 
	}
/*
	cout << cnt++ << " " << hmm.inf(from) << "->" << hmm.inf(to) << "\n"; 
*/
}

int main()
{
	char line[MAXLINE];
	char out_string[9999];
	char temp_string[9999];
	int i, j, k;

	pwt_tf = tdbm_open(PWT_TDBM_FILE, 'r');
	ptt_tf = tdbm_open(PTT_TDBM_FILE, 'r');

  while(1) {
	while(gets(line)!=NULL)
	{
		int is_end_wp;

		if(line[0]==0)
			continue;
		else if(line[0]=='<')
		{
			int v, w;
			char *now_tag;
			int *sp;

			char* tmp_tok;
			/*⿡ head  */
			w=new_wp(line);
			is_end_wp = FALSE;
			sp=&wp[w].mnode;
			while(1)
			{
				if(gets(line)==NULL)
					break;
				if(line[0]==0)
					break;
				
			/*	make sblings */
				now_tag = phtag(line);
				if (strcmp(now_tag, "SF") == 0)
					is_end_wp = TRUE;
				v= new_mnode(line, now_tag, 0.0);
				*sp=v;
				sp=&mn[v].sibling;
			}
		}
		if (is_end_wp) break;
	}

	/*	ͺ .
	*/
	for(i=0;i<wp_end;i++)
	{
		for(j=wp[i].mnode;j!=0;j=mn[j].sibling)
		for(k=wp[i+1].mnode;k!=0;k=mn[k].sibling)
		{
			UpdateProbScore(j,k);
		}
	}
	
	/*  Ѵ. */
	out_string[0] = '\n';
	for(k=wp[i-1].mnode;k!=0;k=mn[k].backptr)
	{
		strcpy(temp_string, out_string);
		sprintf(out_string, "%s\n", mn[k].mout_str, temp_string);
	}
	printf("%s", out_string);
		/*  Ǿ ִ   ϱ ؼ */

	/*  ̻    ´. */
	if ( line == NULL) break;
  }
  tdbm_close(pwt_tf);
  tdbm_close(ptt_tf);
}

