/*---------------------------------------------------------------------- 
	File Name	: trie.h
	compiler	: gcc
	Edited by	: Lee, WoonJae
	Last Change	: 1995.8.24.
			  1999.7.29
	trie export ϱ     ִ.
	  ڰ Ƿ     ϴ.
	   ϸ trie.c ٽ compile ؾ Ѵ.
-----------------------------------------------------------------------*/
#ifndef __trie_h
#define __trie_h 1
#include "dcache.h"

#define TRIE_VERSION "TRIE v0.1 (c) wjlee in KAIST\n"

/* ⺻ Ÿ Ÿ 
	N_KEY : trie node  key ǹѴ. 
		1-255(0 NULL̹Ƿ .)
	G_SIZE :  node  ϱ Ͽ Ǵ Ÿ
		child_size, info_size Ǹ
		0-255, 0  Ÿ.
*/
typedef unsigned long IDX;
typedef unsigned char N_KEY;
typedef unsigned char G_SIZE;

/* TRIE node   -- 12byte
	δ trie node (siblings) ѵ  ȴ.
*/
typedef struct TNODE {
	N_KEY	key;
	N_KEY	dummy;		/* not used */
	G_SIZE	info_size;
	G_SIZE	child_size;
	IDX	info_idx;
	IDX	child_idx;
} TNODE;

/* FREE node 
	free node info node trie_node freeų 
	̰ ߿ ϱ Ͽ ϴ .
	sizeof(FREE) == sizeof(TNODE)  == sizeof(INFO)
*/
typedef struct FREE {	
	IDX	size;
	IDX	lchild;
	IDX	rchild;
} FREE;

/*
	trie : trie data base manager ? structure
*/

typedef struct TRIE {
	int	state;
	int	info_type;
	DCACHE	*cache;		/* trie cache */
	int	search_end;	/* ̰ ˻ path ϴ  */
	long	search_idx[256];/*   listingҶ ַ  */
	N_KEY	search_key[256];
} TRIE;

#define TRIE_HEAD_SIZE		(DC_PAGESIZE*sizeof(TNODE))
#define TRIE_HEAD_SYSTEM		256     /* byte */
#define TRIE_HEAD_USER		(TRIE_HEAD_SIZE-TRIE_HEAD_SYSTEM)

/* state */
#define TRIE_STATE_NEW	1
#define TRIE_STATE_OLD	0

/* info_type */
#define ITYPE_TDBM 	0x00
#define ITYPE_MDIC	0x10
#define ITYPE_TX	0x20
#define ITYPE_TY	0x21

/* ⺻() Լ  */
extern int trie_firstkey(TRIE *tf, N_KEY *str);	/* str is result space */
extern int trie_lastkey(TRIE *tf, N_KEY *str);	/* str is result space */
extern int trie_nextkey(TRIE *tf, N_KEY *str);	/* str is result space */
extern int trie_prevkey(TRIE *tf, N_KEY *str);	/* str is result space */
extern void trie_write_user_head(TRIE *tf, char *thead);
extern void trie_write_user_head(TRIE *tf, char *thead);
extern TRIE *trie_pack(TRIE *tf); 	/*  ʴ.  */

/*----------------------------------------------------------------------
     ̴̽.
  Ʒ Լ ⺻Լ ̿ Ȯ Լ̴.
----------------------------------------------------------------------*/
/* INFO node  -- 12byte
	info node      ִ.
	  ܾ  ǹ̰  
	 ܾ   ٸ  ִ.
	ǻ_1:doctor
	ǻ_2:ǰ Ǵ , 
	ǻ_3:߱ǻ

INFO node δ  Ȯ ʾҴ.
ϸ ¼ ؼ ܰ迡 ʿ   Ϸ Ѵ.
  ٲ  α׷ ٽ  ؾ Ѵ.
׸  text file  ٽ  Ѵ.
*/
typedef struct INFO {
	unsigned long	Word_ID;	/* ܾ ȣ */
	unsigned char	mtag;
	unsigned char	stag;
	unsigned char	phone;		/* p_irr */
	unsigned char	gram;		/* ʿ */
	unsigned char	sem;
	unsigned char	domain;
	unsigned char	wtype; 		/* n s r a */
	unsigned char	dummy;
} INFO;
/* wtype
	n : normal
	s : stop word
	r : ready made
	a : abb.
*/

/* MDIC  : ¼ м  ̴. */
typedef TRIE MDIC;

/* type 'r' ƴϸ 'w' ̺κ DBM  ٸ.  */
extern MDIC *mdic_open(char *fname, char type);
extern void mdic_close(MDIC *tf); 

extern int mdic_store(MDIC *tf, N_KEY *str, INFO *iptr);
extern int mdic_fetch(MDIC *tf, N_KEY *str, INFO *inode);
extern int mdic_lookup(MDIC *tf, N_KEY *str, TNODE tpath[], int tp_size);
		/* tpath[] is result space, tp_size is size of tpath[] */
extern int mdic_delete(MDIC *tf, N_KEY *str, INFO *ip);
/* mdic_lookup  Ѵ. trie.c getinode()  */
#define mdic_info(tf,idx) ((INFO *)dc_add_r((tf)->cache,(idx)+DC_PAGESIZE))

/*  TDBM  */
#define TDBM_INSERT  0
#define TDBM_REPLACE 1

typedef struct info_head {
	unsigned long	Word_ID;
	unsigned long	size;
	unsigned long	info;
} info_head;

typedef TRIE TDBM;
extern TDBM * tdbm_open(char *fname, char type); /* type 'r','w' */
extern void tdbm_close(TDBM *tf);
extern int tdbm_store(TDBM *tf, N_KEY *word, N_KEY *data, int store_mode);
extern int tdbm_fetch(TDBM *tf, N_KEY *str, N_KEY *data);
extern int tdbm_delete(TDBM *tf, N_KEY *str);

/*  TX, TY  */
typedef TRIE TX;
typedef struct txtype {
	unsigned long	next;
	unsigned long	posx;
	unsigned long	posy;
} txtype;

extern TX * tx_open(char *fname, char type); /* type 'r','w' */
extern void tx_close(TX *tf);
extern int tx_store(TX *tf, N_KEY *word, txtype *data);
extern int tx_fetch(TX *tf, N_KEY *str, txtype *data);
extern int tx_count(TX *tf, N_KEY *str);
extern int tx_delete(TX *tf, N_KEY *str, txtype *data);

typedef TRIE TY;
extern TY * ty_open(char *fname, char type); /* type 'r','w' */
extern void ty_close(TY *tf);
extern int ty_store(TY *tf, N_KEY *word, char *data, int ysize, int plen);
extern int ty_fetch(TY *tf, N_KEY *str, char *data);
extern int ty_count(TY *tf, N_KEY *str);

/* debugging */
/*#define DEBUG_NODE_ALLOC	/* node alloc Ȳ  */
/*#define DEBUG_NODE_TRACE	/* node ֱ   */
/*#define DEBUG_LOOK		/* mdic_lookup count */

extern int do_trace(TRIE *tf);
extern void prn_free(FREE *fnode);
extern void prn_tnode(TNODE *tnode);
extern void prn_info(INFO *inode);

#endif

